
#include "Common.h"

/*
 * Fast mutex used to protect the gethostbyname function.
 * It is initialized statically.
 */
pthread_mutex_t host_mutex = PTHREAD_MUTEX_INITIALIZER;
 
/*
 * print_posix_error:
 *   Prints an error message using a supplied error code.
 *   Does not exit the process.
 */
void print_posix_error(int code, const char *msg) 
{
    fprintf(stderr, "%s: %s\n", msg, strerror(code));
}

/*
 * print_unix_error:
 *   Prints an error message using 'errno'. 
 *   'errno' is thread-safe according to the MAN pages.
 *   Does not exit the process.
 */
void print_unix_error(const char *msg) 
{
    fprintf(stderr, "%s: %s\n", msg, strerror(errno));
}

/*
 * print_app_error:
 *   Prints an app error message. Does not exit the process.
 */
void print_app_error(const char *msg) 
{
    fprintf(stderr, "%s\n", msg);
}

/*
 * check_posix_error:
 *   Check whether there's an error and if there is, print an error message.
 * Params:
 *   result:
 *     Result of a function to check for error.
 *   msg:
 *     The error message which will be printed only when there's an error.
 * Return:
 *   Non-negative if there's no error; Negative value otherwise.
 */
int check_posix_error(int result, const char *msg)
{
	if (result < 0)
	{
		print_posix_error(result, msg);
	}
	return result;
}

/*
 * check_unix_error:
 *   Check whether there's an error and if there is, print an error message.
 * Params:
 *   result:
 *     Result of a function to check for error.
 *   msg:
 *     The error message which will be printed only when there's an error.
 * Return:
 *   Non-negative if there's no error; Negative value otherwise.
 */
int check_unix_error(int result, const char *msg)
{
	if (result < 0)
	{
		print_unix_error(msg);
	}
	return result;
}


/*
 * get_host_addr:
 *   Gets the IP address of a host.
 * Params:
 *   name:
 *     Name or decimal dotted numeric IP address of the host.
 *   addr:
 *     The IP address will be copied to the struct pointed to by this.
 * Return:
 *   Non-negative on success; negative otherwise.
 * Notes:
 *   This is thread-safe.
 */
int get_host_addr(const char *name, struct in_addr *addr)
{
	struct hostent *entry;
	int result;
	
	result = check_posix_error(
		pthread_mutex_lock(&host_mutex),
		"ERROR: Unable to lock the host mutex"
	);
	
	if (result >= 0)
	{
		entry = gethostbyname(name);
		if (entry)
		{
			/* Copy over the IP address */
			memcpy(addr, entry->h_addr_list[0], entry->h_length);
			result = 0;
		}
		else
		{
			printf("ERROR: Unable to get host: %s, error=%d\n", 
				name, h_errno);
			result = -1;
		}
	
		/* There's nothing we can do if there's an error except to
		print an error message. */
		check_posix_error(
			pthread_mutex_unlock(&host_mutex),
			"ERROR: Unable to release the host mutex"
		);
	}
	return result;
}


int open_host_fd(const char *host, int port)
{
	int server_fd;
    struct sockaddr_in server_addr;
	
	memset(&server_addr, 0, sizeof(server_addr));
	if (get_host_addr(host, &server_addr.sin_addr) < 0)
	{
		print_app_error("ERROR: Unable to get host addr");
		return -1;
	}

	if ((server_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		print_unix_error("ERROR: Unable to open socket");
		return -1;
	}

	server_addr.sin_family = AF_INET;
	server_addr.sin_port = htons(port);
	
    /* Establish a connection with the server */
    if (connect(server_fd, (struct sockaddr*)&server_addr, 
		sizeof(server_addr)) < 0)
	{
		print_unix_error("ERROR: Unable to connect to host");
		check_unix_error(close(server_fd), 
			"ERROR: Unable to close server socket");
		return -1;
	}
	
	return server_fd;
}


/*
 * sio_init_read:
 *   Initialize an sio_t structure and associate it with a file descriptor.
 * Params:
 *   sio:
 *     Pointer to a sio_t struct to be initialized.
 *   fd:
 *     Descriptor of the socket to be read from.
 * Notes:
 *   Adapted from the RIO code.
 */
void sio_init_read(sio_t *sio, int fd)
{
	sio->fd = fd;  
    sio->num_unread = 0;  
    sio->unread_idx = 0;
}

/*
 * sio_write:
 *   Writes data to a socket.
 * Params:
 *   fd:
 *     Descriptor of the socket to write to.
 *   buf:
 *     Pointer to the data to be written.
 *   n:
 *     Number of bytes to write.
 * Return:
 *   Number of bytes written or negative if there's an error.
 * Notes:
 *   Adapted from the RIO code.
 */
ssize_t sio_write(int fd, const void *buf, size_t n)
{
	size_t num_left = n;
    ssize_t num_written;
    char *cur_ptr = (char*)buf;

    while (num_left > 0) 
	{
		num_written = send(fd, cur_ptr, num_left, MSG_NOSIGNAL);
		if (num_written <= 0)
		{
			/* Were we interrupted by any signal handlers? */
			if (errno == EINTR)
			{
				/* Yes, we are. So we shall call send() again. */
				continue;
			}

			/* There's really an error. */
			return -1;
		}
		num_left -= num_written;
		cur_ptr += num_written;
    }
    return n;
}

/*
 * sio_read_helper:
 *   Helper function for reading of socket data.
 * Params:
 *   sio:
 *     Pointer to an initialized sio_t struct.
 *   buf:
 *     Pointer to the buffer in which data will be stored.
 *   n:
 *     Number of bytes to read.
 * Return:
 *   Returns the actual number of bytes read;
 *   Zero if end-of-file is encountered;
 *   negative if there's an error.
 * Notes:
 *   Adapted from the RIO code.
 */
ssize_t sio_read_helper(sio_t *sio, void *buf, size_t n)
{
    size_t num_to_copy;
	ssize_t num_read;

	/* Refill our buffer while it is empty */
    while (sio->num_unread <= 0) 
	{
		num_read = recv(sio->fd, sio->buffer, sizeof(sio->buffer), 0);
		if (num_read < 0)
		{
			/* Error encountered? */
			if (num_read != EINTR)
			{
				return -1;
			}
		}
		else if (num_read == 0)
		{
			/* End of file encountered. */
			return 0;
		}
		else 
		{
			sio->unread_idx = 0;
			sio->num_unread = (size_t)num_read;
		}
    }

    /* Copy min(n, sio->num_unread) bytes from internal buf to user buf */
    num_to_copy = n;          
    if (sio->num_unread < n)  
	{
		num_to_copy = sio->num_unread;
	}
	
    memcpy(buf, &sio->buffer[sio->unread_idx], num_to_copy);
    sio->num_unread -= num_to_copy;
    sio->unread_idx += num_to_copy;
    return num_to_copy;
}

/*
 * sio_read:
 *   Basically the same as sio_read_helper.
 */
ssize_t	sio_read(sio_t *sio, void *buf, size_t n)
{
    size_t num_left = n;
    ssize_t num_read;
    char *cur_ptr = (char*)buf;
    
    while (num_left > 0) 
	{
		num_read = sio_read_helper(sio, cur_ptr, num_left);
		if (num_read < 0) 
		{
			/* Were we interupted by a sig handler? */
			if (errno == EINTR) 
			{
				/* Yes, keep trying. */
				continue;
			}
			/* There's really an error */
			return -1;
		} 
		else if (num_read == 0)
		{
			/* End of file encountered */
			break; 
		}
		
		num_left -= num_read;
		cur_ptr += num_read;
    }
	/* Should be non-negative */
    return (n - num_left);
}

/* 
 * sio_read_line:
 *   Reads a line-feed terminated line from a socket.
 * Params:
 *   sio:
 *     Pointer to an initialized sio_t struct.
 *   buf:
 *     Pointer to buffer in which to store the data read.
 *   buf_size:
 *     Capacity of the buffer.
 * Return:
 *   Returns the number of characters read, excluding the null terminator.
 *   Or negative if there's an error.
 *   Or zero if end-of-file is reached.
 * Notes:
 *   Adapted from the RIO code.
 */
ssize_t	sio_read_line(sio_t *sio, void *buf, size_t buf_size)
{
	ssize_t num_read;
    char ch;
	char *cur_ptr = (char*)buf;
	char *boundary_end = (char*)buf + buf_size;

	if (buf_size == 0)
	{
		return 0;
	}
	
	/* It's < boundary_end - 1 to ensure we always
	have space for the null terminator. */
    while (cur_ptr < boundary_end - 1)
	{ 
		num_read = sio_read(sio, &ch, 1);
		if (num_read == 1) 
		{
			*cur_ptr = ch;
			cur_ptr++;
			if (ch == '\n')
			{
				break;
			}
		} 
		/* Is it End of file? */
		else if (num_read == 0) 
		{
			break;
		} 
		else
		{	
			/* There was an error. */
			return -1;
		}
	}
	
	*cur_ptr = 0;
	/* Returns the no. of chars, excluding the null char */
    return (ssize_t)(cur_ptr - (char*)buf);
}


