//
//CoordPacket.java
//
//CoordPacket [[int size] [int seq] [float coord0_x][float coord0_y][float coord0_z]...
//          ......................[float coord1_x][float coord1_y][float coord1_z]...
//           ............[float coordsize_x][float coordsize_y][float coordsize_z]]            
//Each CoordPacket will contain <size> as a "header" to allow server or client to know
//how many bytes of data to read in, in readBytes() method. <controllerId> will be for 
//knowing which drawing tool/phone is using the program. <seq> will be useful for detailing
//the sequence of points that are sent. <arr[]> will contain the coordinates (x,y,z) of all the 
//points in the packet, as illustrated above.
//
//Created by Yumin Wong on 2/27/11.
//Last modified by Yumin Wong on 4/22/11.
//

//package edu.cmu.ece.drawe3.comm;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.*;

public class CoordPacket 
 {
  // size will be the number of points (x,y,z) that client wants to create.
  private int size;
  private int controllerId;
  private int seq;
  private int[] arr;
  
  public CoordPacket(int size)
  {
   this.arr = new int [size * 3];
   this.size = size;
  }
  
  /**
   * Creates a new object from a linked list of Coords (inner class below)
   * @param list The linked list
   * @return The coordinate packet created
   */
  public static CoordPacket fromLinkedList (LinkedList<Coord> list, int conId, int seq)
  {
   CoordPacket cp = new CoordPacket(list.size());
   
   cp.controllerId = conId;
   cp.seq = seq;
   cp.arr = new int[cp.size*3];
   for(int i = 0; i < cp.size; i++)
   {
    Coord c = list.poll();
    cp.setPoint(i, c.point);
   }
   return cp;
  }
  
  /**
   * Creates a new object from a byte array
   * @param b The byte array
   * @return The coordinate packet created
   * @throws IOException 
   */
  public static CoordPacket fromByteStream (DataInput in) throws IOException
  {
   CoordPacket cp = new CoordPacket(in.readInt());
   
   cp.controllerId = in.readInt();
   cp.seq = in.readInt();
   cp.arr = new int[3*cp.size];
   
   for(int i = 0; i < cp.size; i++)
   {
    cp.arr[3*i] = in.readInt();
    cp.arr[3*i+1] = in.readInt();
    cp.arr[3*i+2] = in.readInt();
   }
   
   return cp;
  }
  
  /**
   * Returns the byte array representation of the packet
   * @throws IOException 
   */
  public void toByteStream(DataOutput out) throws IOException
  {
   out.writeInt(this.size);
   out.writeInt(this.controllerId);
   out.writeInt(this.seq);
   
   for(int i = 0; i < size; i++)
   {
    out.writeInt(arr[3*i]);
    out.writeInt(arr[3*i+1]);
    out.writeInt(arr[3*i+2]);
   }
  }
  
  /**
   * Returns the string representation of the packet
   * @return The string representation
   */
  public String toString()
  {
   StringBuffer sb = new StringBuffer();
   sb.append("size: "+size);
   sb.append(", controllerId: "+controllerId);
   sb.append(", seq: "+seq);
   for(int i = 0; i < size; i++)
   {
    sb.append(", coordx of "+i+": "+arr[3*i]+" ");
    sb.append(", coordy of "+i+": "+arr[3*i+1]+" ");
    sb.append(", coordz of "+i+": "+arr[3*i+2]+" ");
   }
   return sb.toString();
  }
  
  // Getter methods
  public int getSize()           { return this.size; }
  public int getControllerId() { return this.controllerId; }
  public int getSeq()             { return this.seq; }
  public int[] getPoint (int index)
  {
   if (index < 0 || index > size) return null;
   return new int[] { arr[index * 3], arr[index * 3 + 1], arr[index * 3 + 2]};
   
  }
  
  // Setter methods
  //public void setSize(int size)         { this.size = size; }
  public void setControllerId(int id)  { this.controllerId = id; }
  public void setSeq(int seq)             { this.seq = seq; }
  public void setPoint (int index, int[] point)
  {
   if (index < 0 || index > size) return;
   if (point.length != 3) return;
   arr[index * 3] = point[0];
   arr[index * 3 + 1] = point[1];
   arr[index * 3 + 2] = point[2];
  }
  
  /*
   * An inner class that creates one object that details one point's 3D coordinates
   * Object is created based on a float array of 3 coordinates that user specifies.
   */
  public static class Coord
   {
    int[] point;
    public Coord (int[] p)
    {
     point = p;
    }
   }
 }
